import json
from datetime import date
from cenotes.models import Note
from cenotes_lib import crypto


def assert_successful_request(response):
    assert response.status_code == 200
    assert response.json["success"] is True


def assert_bad_request(response):
    assert response.status_code == 400
    assert response.json["success"] is False
    assert response.json["error"] == "Invalid key or note not found"


def test_encrypt_simple(app, db, client):
    plaintext = "test-note"
    assert Note.query.count() == 0
    response = client.post(
        "notes/encrypt/", data=json.dumps(
            dict(plaintext=plaintext, expiration_date="19-10-2017")),
        content_type='application/json')
    assert_successful_request(response)
    assert Note.query.count() == 1

    note = Note.query.one()
    key = crypto.url_safe_decode(response.json["key"])
    assert crypto.decrypt_with_key(crypto.url_safe_decode(note.payload),
                                   key).decode() == plaintext
    assert str(note.id) == crypto.decrypt_with_box(
        crypto.url_safe_decode(
            response.json["payload"]), app.server_box).decode()
    assert note.expiration_date == date(year=2017, month=10, day=19)


def test_encrypt_no_plaintext(db, client):
    assert Note.query.count() == 0
    response = client.post(
        "notes/encrypt/", data=json.dumps(
            dict(expiration_date="19")),
        content_type='application/json')
    assert response.status_code == 200
    assert Note.query.count() == 1


def test_encrypt_no_note(db, client):

    assert Note.query.count() == 0
    response = client.post("notes/encrypt/",
                           data=json.dumps(dict(key="test")),
                           content_type='application/json')
    assert_successful_request(response)
    note = Note.query.one()
    key = crypto.url_safe_decode(response.json["key"])
    assert crypto.decrypt_with_key(crypto.url_safe_decode(note.payload),
                                   key).decode() == ""


def test_encrypt_no_store(db, client):
    plaintext = "test-note"
    assert Note.query.count() == 0
    response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext, no_store=True)),
        content_type='application/json')
    assert_successful_request(response)
    assert Note.query.count() == 0

    note = crypto.url_safe_decode(response.json["payload"])
    key = crypto.url_safe_decode(response.json["key"])
    assert crypto.decrypt_with_key(note, key).decode() == plaintext


def test_decrypt_payload(client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext, no_store=True)),
        content_type='application/json')
    note = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(note, key))
    assert_successful_request(dec_response)
    assert dec_response.json["plaintext"] == plaintext


def test_decrypt_payload_in_json(client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext, no_store=True)),
        content_type='application/json')
    note = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.post(
        "/notes/", data=json.dumps(dict(payload=note, key=key)),
        content_type='application/json')
    assert_successful_request(dec_response)
    assert dec_response.json["plaintext"] == plaintext


def test_decrypt_payload_wrong_password(client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext, no_store=True)),
        content_type='application/json')
    note = enc_response.json["payload"]
    key = crypto.url_safe_encode(enc_response.json["key"])

    dec_response = client.get("/notes/{0}/{1}".format(note, key))
    assert_bad_request(dec_response)


def test_decrypt_note(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext)),
        content_type='application/json')
    encr_note_id = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_successful_request(dec_response)
    assert dec_response.json["plaintext"] == plaintext


def test_decrypt_note_wrong_password(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext)),
        content_type='application/json')
    encr_note_id = enc_response.json["payload"]
    key = crypto.url_safe_encode(enc_response.json["key"])

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_bad_request(dec_response)


def test_decrypt_note_one_time(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(dict(plaintext=plaintext)),
        content_type='application/json')
    assert Note.query.count() == 1
    encr_note_id = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_successful_request(dec_response)
    assert Note.query.count() == 0


def test_decrypt_note_two_times(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=2)),
        content_type='application/json')
    assert Note.query.count() == 1
    encr_note_id = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_successful_request(dec_response)
    assert Note.query.count() == 1

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_successful_request(dec_response)
    assert Note.query.count() == 0


def test_decrypt_note_more_than_allowed(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=1)),
        content_type='application/json')
    assert Note.query.count() == 1
    encr_note_id = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_successful_request(dec_response)
    assert Note.query.count() == 0

    dec_response = client.get("/notes/{0}/{1}".format(encr_note_id, key))
    assert_bad_request(dec_response)


def test_delete_note_in_json(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=10)),
        content_type='application/json')
    note = enc_response.json["payload"]
    duress_key = enc_response.json["duress_key"]

    assert Note.query.count() == 1

    dec_response = client.post(
        "/notes/", data=json.dumps(dict(payload=note, key=duress_key)),
        content_type='application/json')
    assert_bad_request(dec_response)
    assert Note.query.count() == 0


def test_delete_note(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=10)),
        content_type='application/json')
    note = enc_response.json["payload"]
    duress_key = enc_response.json["duress_key"]

    assert Note.query.count() == 1

    assert_bad_request(client.get("/notes/{0}/{1}".format(note, duress_key)))
    assert Note.query.count() == 0


def test_delete_note_multiple_times(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=10)),
        content_type='application/json')
    note = enc_response.json["payload"]
    duress_key = enc_response.json["duress_key"]

    assert_bad_request(client.get("/notes/{0}/{1}".format(note, duress_key)))
    assert_bad_request(client.get("/notes/{0}/{1}".format(note, duress_key)))


def test_avoid_note_deletion_issuing_payload_as_key(db, client):
    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/",
        data=json.dumps(dict(plaintext=plaintext, max_visits=10)),
        content_type='application/json')
    note = enc_response.json["payload"]

    assert Note.query.count() == 1

    dec_response = client.get("/notes/{0}/{1}".format(note, note))

    assert_bad_request(dec_response)
    assert Note.query.count() == 1


def test_get_config(client):
    response = client.get("config/algorithms/")

    assert response.status_code == 200
    # scrypt is the default fallback option, so it should always exist
    assert response.json["scrypt"] is not None


def test_get_default_algorithm_options(client):
    response = client.get("config/algorithms/default/")

    assert response.status_code == 200
    assert response.json["algorithm"] == "scrypt"
    assert response.json["hardness"] == "min"


def test_encrypt_with_options(app, client):
    algo, hardness = "scrypt", "moderate"
    assert hardness != app.config["FALLBACK_ALGORITHM_PARAMS"]["hardness"]

    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(
            dict(plaintext=plaintext, no_store=True,
                 algorithm=algo, hardness=hardness)),
        content_type='application/json')
    note = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(note, key))
    assert_successful_request(dec_response)
    assert dec_response.json["plaintext"] == plaintext


def test_encrypt_with_wrong_options(app, client):
    algo, hardness = "wrong", "wrong"
    assert hardness != app.config["FALLBACK_ALGORITHM_PARAMS"]["hardness"]

    plaintext = "test-note"
    enc_response = client.post(
        "notes/encrypt/", data=json.dumps(
            dict(plaintext=plaintext, no_store=True,
                 algorithm=algo, hardness=hardness)),
        content_type='application/json')
    note = enc_response.json["payload"]
    key = enc_response.json["key"]

    dec_response = client.get("/notes/{0}/{1}".format(note, key))
    assert_successful_request(dec_response)
    assert dec_response.json["plaintext"] == plaintext



from datetime import date, timedelta
import pytest
from cenotes import create_app, db as _db
from cenotes.config_backend import Testing
from cenotes.models import Note
from cenotes_lib import crypto


@pytest.fixture(name="app", scope="session")
def _app():
    return create_app(app_settings=Testing)


@pytest.fixture(scope="function")
def db(app):
    _db.drop_all()
    _db.app = app
    _db.create_all()
    yield _db
    _db.drop_all()


@pytest.fixture(scope="session", name="testing_key")
def _key():
    return crypto.craft_key_from_password("testing")


@pytest.fixture(scope="session", name="testing_box")
def _box(testing_key):
    return crypto.craft_secret_box(testing_key)


@pytest.fixture(scope="function", name="note")
def _note(db):
    note = Note(b"test")
    db.session.add(note)
    db.session.commit()
    return note


@pytest.fixture(scope="function", name="old_note")
def _old_note(db):
    note = Note(b"test", expiration_date=date.today() - timedelta(weeks=36))
    db.session.add(note)
    db.session.commit()
    return note



from flask import jsonify, current_app
from cenotes_lib.exceptions import InvalidUsage


def craft_response(enote, error, key, dkey, payload, plaintext, success):
    return {"success": success if not error else False,
            "error": error,
            "key": key,
            "duress_key": dkey,
            "plaintext": plaintext,
            "payload": payload,
            "expiration_date": enote.iso_expiration_date if enote else "",
            "max_visits": enote.max_visits if enote else ""}


def craft_json_response(success=True, error="", enote=None,
                        plaintext="", key="", payload="", dkey=""):

    return jsonify(
        craft_response(enote, error, key, dkey, payload, plaintext, success))


def get_request_params(request_params):
    try:
        return CENParams(**request_params)
    except TypeError:
        raise InvalidUsage("Wrong format of parameters given. Could not parse")


class CENParams(object):
    def __init__(self, plaintext="", key=None, expiration_date=None,
                 visits_count=None, max_visits=None, no_store=False,
                 payload=None, algorithm=None, hardness=None,
                 **kwargs):
        self.plaintext = plaintext
        self.key = key
        self.payload = payload
        self.expiration_date = expiration_date
        self.visits_count = visits_count
        self.max_visits = max_visits
        self.no_store = no_store
        self.algorithm, self.hardness = self.enforce_correct_encr_options(
            algorithm, hardness)

    @staticmethod
    def enforce_correct_encr_options(algorithm, hardness):
        supported = current_app.config["SUPPORTED_ALGORITHM_PARAMS"]
        fallback = current_app.config["FALLBACK_ALGORITHM_PARAMS"]

        matched_algorithm = (algorithm if algorithm in supported.keys()
                             else fallback["algorithm"])

        if hardness not in supported[matched_algorithm]["hardness"]:
            return tuple(map(
                lambda x: current_app.config["FALLBACK_ALGORITHM_PARAMS"][x],
                ("algorithm", "hardness")))
        else:
            return matched_algorithm, hardness



def test_404(client):
    response = client.get("non_existent")
    assert response.status_code == 404
    assert response.json["success"] is False
    assert response.json["error"] != ""
    assert response.json["key"] == ""
    assert response.json["payload"] == ""
    assert response.json["plaintext"] == ""
    assert response.json["expiration_date"] == ""
    assert response.json["max_visits"] == ""


def test_400(client):
    response = client.get("/notes/encrypt/")
    assert response.status_code == 400
    assert response.json["success"] is False
    assert response.json["error"] != ""
    assert response.json["key"] == ""
    assert response.json["payload"] == ""
    assert response.json["plaintext"] == ""
    assert response.json["expiration_date"] == ""
    assert response.json["max_visits"] == ""



import os
import binascii
import json
from collections import namedtuple
from cenotes_lib import crypto, exceptions

AlgoParam = namedtuple("AlgoParam", ['algorithm', 'hardness'])


def validate_algorithm_params(params):
    supported = crypto.SUPPORTED_ALGORITHM_PARAMS

    def verify_options(algo):
        return (algo.algorithm,
                tuple(filter(lambda x: supported.get(algo.algorithm, {}).get(
                    'hardness', {}).get(x) is None,
                             algo.hardness)))

    return tuple(filter(lambda x: x[-1] != (), map(verify_options, params)))


def is_url_base64_encoded(what):
    try:
        crypto.url_safe_decode(what)
    except (binascii.Error, exceptions.CenotesError):
        return False
    else:
        return True


class Config(object):
    DEBUG = DEVELOPMENT = TESTING = False

    # Define the application directory
    BASE_DIR = os.path.abspath(os.path.dirname(__file__))

    # Define the database - we are working with
    # SQLite for this example
    SQLALCHEMY_DATABASE_URI = os.getenv("DB_URI", 'sqlite:///{0}'.format(
        os.path.join(BASE_DIR, 'cenotes.sqlite3')))
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    DATABASE_CONNECT_OPTIONS = {}

    # Enable protection against *Cross-site Request Forgery (CSRF)*
    CSRF_ENABLED = True

    # Use a secure, unique and absolutely secret key for
    # signing the data.
    CSRF_SESSION_KEY = os.getenv("CSRF_KEY", crypto.generate_random_chars())

    # Secret key for signing cookies
    SECRET_KEY = os.getenv("SECRET_KEY", crypto.generate_random_chars())

    # Server key must be 32-bytes long
    SERVER_ENCRYPTION_KEY = os.getenv("SERVER_ENCRYPTION_KEY", "").encode()
    if is_url_base64_encoded(SERVER_ENCRYPTION_KEY):
        SERVER_ENCRYPTION_KEY = crypto.url_safe_decode(SERVER_ENCRYPTION_KEY)

    if SERVER_ENCRYPTION_KEY and len(SERVER_ENCRYPTION_KEY) != 32:
        raise RuntimeError("Key must be 32 bytes long\n"
                           "Leave it empty if you want to autogenerate one")

    SUPPORTED_ALGORITHM_PARAMS = (
            json.loads(os.getenv("SUPPORTED_ALGORITHM_PARAMS", '{}'))
            or
            {algo: {"hardness": hardness}
             for algo, hardness in tuple(
                crypto.get_supported_algorithm_options())}
    )

    FALLBACK_ALGORITHM_PARAMS = {
        "algorithm": os.getenv("FALLBACK_ALGORITHM") or "scrypt",
        "hardness": os.getenv("FALLBACK_ALGORITHM_HARDNESS") or "interactive"
    }


class Production(Config):
    DEBUG = DEVELOPMENT = TESTING = False


class Development(Config):
    DEBUG = True
    DEVELOPMENT = True


class Testing(Config):
    TESTING = True
    SQLALCHEMY_DATABASE_URI = os.getenv("TEST_DB_URI", 'sqlite:///{0}'.format(
        os.path.join(Config.BASE_DIR, 'cenotes_test.sqlite3')))

    FALLBACK_ALGORITHM_PARAMS = {
        "algorithm": os.getenv("FALLBACK_ALGORITHM") or "scrypt",
        "hardness": os.getenv("FALLBACK_ALGORITHM_HARDNESS") or "min"
    }


def validate_crypto_configuration(config):
    algorithm_params = config["SUPPORTED_ALGORITHM_PARAMS"]
    if not isinstance(algorithm_params, dict):
        raise RuntimeError("Algorithm params must be of dict type."
                           "See the docs for more info")

    unsupported = validate_algorithm_params(
        map(lambda x: AlgoParam(x[0], x[1]['hardness']),
            algorithm_params.items()))
    if unsupported:
        raise RuntimeError(
            "The following algorithms/algorithm-options are not supported "
            "in your system.\n {0}\n"
            "Please verify your configuration and try again"
            .format(unsupported))

    fbalgo = config["FALLBACK_ALGORITHM_PARAMS"]["algorithm"]
    fbhard = config["FALLBACK_ALGORITHM_PARAMS"]["hardness"]

    if not (algorithm_params.get(fbalgo)
            and fbhard in algorithm_params.get(fbalgo, {}).get("hardness")):
        raise RuntimeError("Fallback algorithm parameters must exist in the "
                           "defined supported algorithm params")


def validate_config(config):
    try:
        validate_crypto_configuration(config)
    except RuntimeError as e:
        print("Error: {}".format(e))
        exit(1)



from cenotes.api import CENParams


def test_wrong_algorithm_keeps_hardness(app):
    algo, hardness = CENParams.enforce_correct_encr_options(
        algorithm="lalala", hardness="sensitive")
    assert algo == "scrypt"
    assert hardness == "sensitive"


def test_wrong_algorithm_hardness(app):
    algo, hardness = CENParams.enforce_correct_encr_options(
        algorithm="scrypt", hardness="se")
    assert algo == "scrypt"
    assert hardness == "min"


def test_no_algorithm_options_given(app):
    params = CENParams()
    assert params.algorithm == app.config[
        "FALLBACK_ALGORITHM_PARAMS"]["algorithm"]
    assert params.hardness == app.config[
        "FALLBACK_ALGORITHM_PARAMS"]["hardness"]



from datetime import date, timedelta
import pytest
from sqlalchemy.orm.exc import NoResultFound
from cenotes import models


def test_note_has_expired(old_note):
    assert old_note.has_expired


def test_note_too_many_visits(db, note):
    note.max_visits = 0
    db.session.add(note)
    db.session.commit()
    assert note.has_expired is False
    assert note.should_not_exist


def test_fetch_note(note):
    assert models.fetch_note(note.id) is not None


def test_fetch_expired_note(db, note):
    note.expiration_date = date.today() - timedelta(weeks=1)
    db.session.add(note)
    db.session.commit()
    with pytest.raises(NoResultFound):
        models.fetch_note(note.id)


def test_update_note(db, note):
    note.max_visits = 2
    db.session.add(note)
    db.session.commit()
    assert note.visits_count == 0
    models.update_note(note)
    assert note.visits_count == 1


def test_update_note_delete(note):
    assert models.Note.query.count() == 1
    models.update_note(note)
    assert models.Note.query.count() == 0


def test_fetch_note_paranoid(note):
    assert models.Note.query.count() == 1
    models.fetch_note(note.id)
    assert models.Note.query.count() == 0
